// This program illustrates the calculations of the semiparametric fitting method
// described in Rolke and Lopez "Estimating a Signal In the Presence of an Unknown Background"
//
// The data can be d-dimensional with the boundaries of the  form Ai<xi<Bi, i=1,..,d
// The data is transformed and then the non-parametric density estimate is found using a Gaussian kernel
// The KDE can be found directly (DoDirect=1) or using a Inverse Gaussian Transform algorithm called
// figtree, which is much faster. For details see www.umiacs.umd.edu/~morariu/figtree
//
// if the bandwidth is not specified (h1=0) it is found via the nrd "rule-of-thump" method by Scott (1992)
// independently for each dimension.
//
// The signal is assumed to be a multivariate normal with means mui and standard deviations sigma i and
// correlations 0. This can easily be changed in the routone dsignal.
//
// The mle of the number of signal events is found via  Newton-Raphson in find.mle. If fitting for
// other parameters is desired this can be changed easily.
//
// if desired (nBoot>0) error estimates are found via the statistical bootstrap with nBoot replications
//
// as an example the included file setup.dat has 1000 events in the pure background
// and 1000 events in the data sample and d=2. The signal events are from a N(3,0.15) and a N(1,0.05)
// in the two dimensions
// The true number of signal events is 50.
// The output should be 
// Optimal bandwidths: 0.454  0.424
// Estimated number of signal events:  41.310
// Error:  9.614
 

#include <math.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "figtree.h"

#ifdef WIN32
  // Link to figtree.dll if WIN32.  If not, then we assume that 
  // libfigtree.so or libfigtree.a are linked to from the Makefile.
  // The locations of figtree.dll and ANN.dll must either be in the PATH
  // environment variable, or in the same location from which sample.exe is
  // is executed.
  #pragma comment(lib,"../lib/figtree.lib")
#endif

const int dimmax=10;
const double zero=0.0;
const double one=1.0;

double dnorm(double, double, double);
double pnorm(double, double, double);
void nrd(int,int, double [], double []);
void transform(int, int, int, double [], double [], double [], double [], double []); 
void findmle(int, double [], double [], double []);
void dnp(int, int, int, double [],  double [],  double []);
void dsignal(int, int, double [], double [], double [], double [], double [], double []);
void bubblesort (double [], int);
double sd(double [], int);

int main() {
        
     int nx,ny,l,dim,seed,nBoot,DoDirect;
     double epsilon = 1e-4;
     double mean[dimmax],std[dimmax],A[dimmax],B[dimmax],h[dimmax];
     double estimates[3];
     float f; 
    
     //Read data from file
     FILE * infile; 
     
     infile = fopen ("setup.dat","r");
     fscanf (infile, "%d", &ny); // Number of pure Background events
     fscanf (infile, "%d", &nx); //Number of events in signal data set
     fscanf (infile, "%d", &dim); // dimensionality
     fscanf (infile, "%d", &nBoot); // desired number of bootstrap repetitions, none of 0
     fscanf (infile, "%d", &DoDirect); // 0=figtree, 1=direct calculation
     fscanf (infile, "%d", &seed);
     srand(seed);     
     
     for(int i=0;i<dim;++i) {fscanf (infile, "%f", &f);A[i]=f;} // lower limits of data
     for(int i=0;i<dim;++i) {fscanf (infile, "%f", &f);B[i]=f;} // upper limits of data
     for(int i=0;i<dim;++i) {fscanf (infile, "%f", &f);h[i]=f;} // bandwidths, 0 if to be calculated from data
     for(int i=0;i<dim;++i) {fscanf (infile, "%f", &f);mean[i]=f;} // means of Gaussian signal
     for(int i=0;i<dim;++i) {fscanf (infile, "%f", &f);std[i]=f;} // stds of Gaussian signal
     
//   read pure background data, y[i*dim],..,y[(i+1)*dim-1] has ith event      
     double * y = new double[dim*ny];
     for(int i = 0; i < ny; i++ ) {
        for(int j=0;j<dim;++j) {
            fscanf (infile, "%f", &f);
            y[i*dim+j]=f;
        }    
     } 

//   read data with potential signal
     double * x = new double[dim*nx];
     for(int i = 0; i < nx; i++ ) {
        for(int j=0;j<dim;++j) {
           fscanf (infile, "%f", &f);
           x[i*dim+j]=f;
        }
     }   
     fclose (infile);            
     
//    q's are weights of events y, set to 1 here, used in figtree    
      double * q = new double[ny];
      for( int i = 0; i < ny; i++ ) q[i]=1.0;
      
      
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++   
// Find the density for the Gaussian signal part    

     double * fSig = new double[dim*nx];    
     memset(fSig, 0, sizeof(double)*dim*nx);  

     dsignal(nx, dim, mean,std, A, B, x, fSig);   

// Transform the data from [Ai,Bi] to whole real line
     double * dxdy = new double[dim*nx];    
     memset(dxdy, 0, sizeof(double)*dim*nx);  
     transform(ny, nx, dim, A, B, y, x, dxdy);

// Find density of signal in new scale     
     for(int i=0;i<nx;++i) fSig[i]=fSig[i]*dxdy[i];             
          
// Now the KDE
// If needed find optimal bandwidth     
     if(h[0]==0) {
        nrd(ny, dim, h, y);        
        printf("Optimal bandwidths: %.3lf",h[0]);
        if(dim>1) for(int i=1;i<dim;++i) printf("  %.3lf",h[i]);
        printf("\n");
     }    

//Rescale h,y and x so that hi=1    
     double prodh=pow(3.141593,-dim/2.0);
     for(int i = 0;i < dim; ++i) {
        h[i]=sqrt(2.0)*h[i];
        prodh=prodh/h[i];
     }    
     for(int i = 0; i < ny; i++) {
        for(int j=0;j<dim;++j) y[i*dim+j]=y[i*dim+j]/h[j];
     } 
     for(int i = 0; i < nx; i++) {
        for(int j=0;j<dim;++j) x[i*dim+j]=x[i*dim+j]/h[j];
     }     
     
     double * fBack = new double[dim*nx];
     memset(fBack, 0, sizeof(double)*dim*nx );        

     if(DoDirect==1) {
          dnp(ny, nx, dim, y, x, fBack);
     }
     else {
          figtree(dim, ny, nx, 1, y, one, q, x, epsilon, fBack);
     }     
//   if fBack=0 set to very small number, scale fBack to be density  
     for(int i=0;i<nx;++i) {
          if(fBack[i]==0) fBack[i]=epsilon/20.0;
          else fBack[i]=fBack[i]*prodh/ny;
     }

// Find the mle of the number of signal events and the error via Newton-Raphson     
     estimates[0]=0.1; // starting point
     findmle(nx, estimates, fBack, fSig); 
    
     printf("Estimated number of signal events:  ");
     printf("%.3lf\n",estimates[0]);
     if(estimates[0]>0) {
        printf("Error:  ");
        printf("%.3lf\n",estimates[1]);
     }
      
// If desired find the error via the Bootstrap
     if(nBoot>0 && estimates[0]>0) { 

//        save data in new arrays       
          double * datay = new double[dim*ny];
          memset(datay, 0, sizeof(double)*dim*ny);
          double * fSigdata = new double[nx];   
          memset(fSigdata, 0, sizeof(double)*dim*nx);      
          double * Bootalpha = new double[nBoot];          
          memset(Bootalpha, 0, sizeof(double)*nBoot);
          for(int i = 0; i < ny; i++) {
              for(int j=0;j<dim;++j) datay[i*dim+j]=y[i*dim+j]*h[j];
          } 
          double * datax = new double[dim*nx];
          for(int i = 0; i < nx; i++) {
              fSigdata[i]=fSig[i];
              for(int j=0;j<dim;++j) datax[i*dim+j]=x[i*dim+j]*h[j];
          }          
          
          for(int k=0;k<nBoot;++k) {
                // find Bootstrap samples          
                for(int i=0;i<ny;++i) {
                    double u=rand()/(double(RAND_MAX)+1);
                    int l=ny*u;
                    for(int j=0;j<dim;++j) y[i*dim+j]=datay[l*dim+j];
                }            
                for(int i=0;i<nx;++i) {
                    double u=rand()/(double(RAND_MAX)+1);
                    int l=nx*u;
                    for(int j=0;j<dim;++j) x[i*dim+j]=datax[l*dim+j];
                    fSig[i]=fSigdata[l];
                }
                // and then do everything as before, but now for Bootstrap sample
                nrd(ny, dim, h, y);
                prodh=pow(3.141593,-dim/2.0);
                for(int j = 0;j < dim; ++j) {
                    h[j]=sqrt(2.0)*h[j];
                    prodh=prodh/h[j];
                }
                
                for(int i = 0; i < ny; i++) {
                    for(int j=0;j<dim;++j) y[i*dim+j]=y[i*dim+j]/h[j];
                } 
                for(int i = 0; i < nx; i++) {
                    for(int j=0;j<dim;++j) x[i*dim+j]=x[i*dim+j]/h[j];
                }                                     
                if(DoDirect==1) {
                      dnp(ny, nx, dim, y, x, fBack);
                }
                else {
                      figtree( dim, nx, ny, 1, y, one, q, x, epsilon, fBack);
                }     
                for(int i=0;i<nx;++i) {
                    if(fBack[i]==0) fBack[i]=epsilon/20.0;
                    else fBack[i]=fBack[i]*prodh/ny;                
                }                
                estimates[0]=0.1;
                findmle(nx, estimates, fBack, fSig);               
                Bootalpha[k]=estimates[0];
          }      
          estimates[1]=sd(Bootalpha,nBoot);
          printf("Bootstrap Error:  ");
          printf("%.3lf\n",estimates[1]); 
          delete [] fSigdata;
          delete [] Bootalpha; 
          delete [] datay;
          delete [] datax;                
     }          
     
     delete [] fSig;
     delete [] fBack;
     delete [] x;
     delete [] y;

     return(0);
}
/***************************************************************************************************/
/***************************************************************************************************/
// Transform data from [Ai,Bi] to R
void transform(int ny, int nx, int dim, double A[], double B[], double y[], double x[], double dxdy[]) {

    for(int i=0;i<ny;++i) 
         for(int j=0;j<dim;++j) y[i*dim+j]=log( (y[i*dim+j]-A[j])/(B[j]-y[i*dim+j]) );
    for(int i=0;i<nx;++i) {
        dxdy[i]=one;    
        for(int j=0;j<dim;++j) {
          x[i*dim+j]=log( (x[i*dim+j]-A[j])/(B[j]-x[i*dim+j]) );
          dxdy[i]=dxdy[i]*(B[j]-A[j])*exp(x[i*dim+j])/pow((exp(x[i*dim+j])+1),2);
        }
    }       
} 
/***************************************************************************************************/
// Normal density
double dnorm(double x, double mu, double sigma) {
    
    double y=0.3989423/sigma*exp(-0.5*(x-mu)*(x-mu)/sigma/sigma);
    return(y);
}     
 
/***************************************************************************************************/
// Parametric Gaussian Signal
void dsignal(int nx, int dim, double mean [], double std [], double A [], double B [], double x [], double fSig []) {

    double cc=1.0;
    for(int i=0;i<dim;++i) cc=cc*(pnorm(B[i],mean[i],std[i])-pnorm(A[i],mean[i],std[i]));
    for(int i=0;i<nx;++i) {
        fSig[i]=1.0/cc;
        for(int j=0;j<dim;++j) {
            fSig[i]=fSig[i]*dnorm(x[i*dim+j],mean[j], std[j]);
        }    
    }
}
/***************************************************************************************************/
// Find standard deviation
double sd(double x[],int n) {

    double sxx,s;
    double tmp[2];

    tmp[0]=0;
    tmp[1]=0;
    for(int i=0;i<n;++i) {
        tmp[0]=tmp[0]+x[i];
        tmp[1]=tmp[1]+x[i]*x[i];
    }
    sxx=tmp[1]-tmp[0]*tmp[0]/n;
    s=sqrt(sxx/(n-1));
    return(s);
}    
/***************************************************************************************************/
// Find optimal bandwith from pure background sample using Scott's "Rule-of-Thumb"
void nrd(int ny, int dim, double h[], double y[]) {

    int l;
    double IQR,s,tmp[2];
    
    double * ytmp = new double[ny];

    for(int j=0;j<dim;++j) {
        for(int i=0;i<ny;++i) ytmp[i]=y[i*dim+j];
        s=sd(ytmp,ny);
        bubblesort (ytmp, ny);
        l=0.25*ny;
        tmp[0]=0.75*ytmp[l]+0.25*ytmp[l+1];
        l=0.75*ny;
        tmp[1]=0.25*ytmp[l-1]+0.75*ytmp[l];
        IQR=(tmp[1]-tmp[0])/1.34;
        if(s<IQR) h[j]=1.06*s*pow(ny,-0.2);
        else h[j]=1.06*IQR*pow(ny,-0.2);
    }
}
/***************************************************************************************************/
// Find non-parametric density estimates at points x using data from y
void dnp(int ny, int nx, int dim, double y[], double x[], double fBack[] ) {
    
    for(int j=0;j<nx;++j) {     
       for(int i=0;i<ny;++i) {
            double tmp=0.0;  
            for(int k=0;k<dim;++k) tmp+=pow(y[i*dim+k]-x[j*dim+k],2);     
            fBack[j] += exp(-tmp);
       }
    }       
}
   /***************************************************************************************************/       
// use Newton-Raphson to find mle of number of signal events
void findmle(int nx, double estimates[], double fBack[], double fSig[]) {      

     double alpha,alphaold,tmp,Numer,Denom;
     double ll[2];
        
     tmp=0; 
     for(int i=0;i<nx;++i) tmp=tmp+(fSig[i]-fBack[i])/fBack[i];                 

     if(tmp>0) {
        alpha=estimates[0];
        do {
            alphaold=alpha;
            Numer=0;
            Denom=0;       
            for(int i=0;i<nx;++i) {
                tmp=(fSig[i]-fBack[i])/((1-alpha)*fBack[i]+alpha*fSig[i]);
                Numer=Numer+tmp;
                Denom=Denom-tmp*tmp;
            }
            double adj=2.0;
            int k=0;
            do {
                ++k;
                adj=adj/2.0;
                alpha=alphaold-adj*Numer/Denom;
                if(k>10) goto Done;
            }
            while (alpha<0);    
        }     
        while (fabs(alpha-alphaold)>pow(10,-7));                        
     
        Done:
        ll[0]=0;
        ll[1]=0;
        for(int i=0;i<nx;++i) {
          ll[0]=ll[0]+log(fBack[i]);
          ll[1]=ll[1]+log((1-alpha)*fBack[i]+alpha*fSig[i]);
        }
        estimates[0]=nx*alpha;
        estimates[1]=nx/sqrt(-Denom);        
        estimates[2]=2*(ll[1]-ll[0]);
    }
    else for(int i=0;i<3;++i) estimates[i]=0;
}
//*************************************************  
// Sort values in x
void bubblesort (double x[], int vec_length)
{
    int j,Sorted,first,last;
    double hold;
    
    Sorted = 0;
    first = 0;
    last = vec_length - 2;
    while(Sorted==0) {
        Sorted = 1;
        for(j=first;j<=last;++j) {
            if (x[j]>x[j+1]) {
                hold = x[j];
                x[j] = x[j+1];
                x[j+1] = hold;
                Sorted = 0;
            }
        }   
        last = last - 1;
    }    
}      
//*************************************************
//Cumulative distribution function of Gaussian
double pnorm(double x, double mu, double sigma) {
    
    static double c=0.398942280401,c1=0.585786437627,pi=3.14159265359;
    double p,z,x2;
    int neg;

    z=x;
    if(z<mu) {
        neg=1;
        x=(mu-z)/sigma;
    }
    else {
        neg=0;
        x=(z-mu)/sigma;
    }        
    x2=pow(x,2);

    if(x<0.5) {
        p=0.5+c*(x-pow(x,3)/6+pow(x,5)/40-pow(x,7)/336+pow(x,9)/3456);
    }    
    else {
        if(x<3) {
            p=0.5+0.5*sqrt(1.0-1.0/30.0*(7.0*exp(-x2/2.0)+16.0*exp(-x2*c1)+(7.0+0.25*pi*x2)*exp(-x2)));
        }
        else {
            p = 1-c*exp(-0.5*x2)*(1.0/x-1.0/pow(x,3)+3.0/pow(x,5)-15.0/pow(x,7)+105.0/pow(x,9));
        }
    }        
    if(neg) p = 1.0-p;
    return(p);
}
